"""Test cache functions."""

# native
from datetime import timedelta
from datetime import datetime

# pkg
from freshlinks.cache import to_timedelta
from freshlinks.cache import age_time
from freshlinks.cache import hash_url
from freshlinks.cache import LinkEntry


LOCALHOST = "http://localhost:8080/"


def test_to_timedelta() -> None:
    """Convert strings to time deltas."""
    assert to_timedelta("1d") == timedelta(days=1)
    assert to_timedelta("2h") == timedelta(hours=2)
    assert to_timedelta("30m") == timedelta(minutes=30)
    assert to_timedelta("45s") == timedelta(seconds=45)

    assert to_timedelta("1d 2h 30m 45s") == timedelta(
        days=1, hours=2, minutes=30, seconds=45
    )
    assert to_timedelta("1d2h30m45s") == timedelta(
        days=1, hours=2, minutes=30, seconds=45
    )

    assert to_timedelta("1.5h") == timedelta(hours=1.5)
    assert to_timedelta("0.5d 12m") == timedelta(days=0.5, minutes=12)

    assert to_timedelta("1x") == timedelta(0)


def test_age_time() -> None:
    """Convert relative durations to a unix timestamp."""
    assert age_time("1s") == int((datetime.now() - timedelta(seconds=1)).timestamp())


def test_hash_url() -> None:
    """Hash URLs."""
    test = LOCALHOST
    want = "387db443878e39df26fe9826db1fd40100b146f93795b07fda3bfeef709b5aca"
    have = hash_url(test)
    assert have == want


def test_load_entry() -> None:
    """Load an entry."""
    test = dict(v=1, t=2000, c=200, err="", url=LOCALHOST, links=[])
    want = LinkEntry(is_cached=True, time=2000, url=LOCALHOST)
    assert LinkEntry.load(test) == want
    assert want.json == test



"""Test `canonical_ip` and `canonical_url` using Google's test suite.

See:
    - https://developers.google.com/safe-browsing/v4/urls-hashing#canonicalization
    - https://chromium.googlesource.com/external/google-safe-browsing/+/06a8c4e799233da220ad7411e2bfacc74cbfbb37/python/expression_test.py
"""

# This file is a partial rewrite of Google's Python 2.5 implementation.
#
# Copyright 2023 Metaist LLC.
# Licensed under the MIT License.
#
# The original `expression_test.py` file carries this license:
#
# Copyright 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# std
from typing import List
from typing import Tuple
from typing import Union

# pkg
from freshlinks.canonicalize import canonical_ip
from freshlinks.canonicalize import canonical_url


def test_canonicalize_ip() -> None:
    """Canonicalize IP addresses."""
    items = [
        ("1.2.3.4", "1.2.3.4"),
        ("012.034.01.055", "10.28.1.45"),
        ("0x12.0x43.0x44.0x01", "18.67.68.1"),
        ("167838211", "10.1.2.3"),
        ("12.0x12.01234", "12.18.2.156"),
        ("0x10000000b", "0.0.0.11"),
        ("asdf.com", ""),
        ("0x120x34", ""),
        ("123.123.0.0.1", ""),
        ("1.2.3.00x0", ""),
        ("fake ip", ""),
        ("123.123.0.0.1", ""),
        ("255.0.0.1", "255.0.0.1"),
        ("12.0x12.01234", "12.18.2.156"),
        # TODO: Make this test case work.
        # This doesn't seem very logical to me, but it might be how microsoft's
        # dns works.  Certainly it's how Netcraft does it.
        # ('276.2.3', '20.2.0.3'),
        ("012.034.01.055", "10.28.1.45"),
        ("0x12.0x43.0x44.0x01", "18.67.68.1"),
        ("167838211", "10.1.2.3"),
        ("3279880203", "195.127.0.11"),
        ("4294967295", "255.255.255.255"),
        ("10.192.95.89 xy", "10.192.95.89"),
        ("1.2.3.00x0", ""),
        # If we find bad octal parse the whole IP as decimal or hex.
        ("012.0xA0.01.089", "12.160.1.89"),
    ]
    for test, want in items:
        have = canonical_ip(test)
        assert have == want, f"test input: {test}, actual:{have}, expected: {want}"


def test_canonicalize_url() -> None:
    """Canonicalize URLs."""
    items: List[Tuple[Union[str, bytes], str]] = [
        ("http://google.com/", "http://google.com/"),
        ("http://google.com:80/a/b", "http://google.com/a/b"),
        ("http://google.com:80/a/b/c/", "http://google.com/a/b/c/"),
        ("http://GOOgle.com", "http://google.com/"),
        ("http://..google..com../", "http://google.com/"),
        ("http://google.com/%25%34%31%25%31%46", "http://google.com/A%1F"),
        ("http://google^.com/", "http://google^.com/"),
        ("http://google.com/1/../2/././", "http://google.com/2/"),
        ("http://google.com/1//2?3//4", "http://google.com/1/2?3//4"),
        # Some more examples of our url lib unittest.
        ("http://host.com/%25%32%35", "http://host.com/%25"),
        ("http://host.com/%25%32%35%25%32%35", "http://host.com/%25%25"),
        ("http://host.com/%2525252525252525", "http://host.com/%25"),
        ("http://host.com/asdf%25%32%35asd", "http://host.com/asdf%25asd"),
        ("http://host.com/%%%25%32%35asd%%", "http://host.com/%25%25%25asd%25%25"),
        ("http://www.google.com/", "http://www.google.com/"),
        (
            "http://%31%36%38%2e%31%38%38%2e%39%39%2e%32%36/%2E%73%65%63%75%72%65/%77%77%77%2E%65%62%61%79%2E%63%6F%6D/",
            "http://168.188.99.26/.secure/www.ebay.com/",
        ),
        (
            "http://195.127.0.11/uploads/%20%20%20%20/.verify/.eBaysecure=updateuserdataxplimnbqmn-xplmvalidateinfoswqpcmlx=hgplmcx/",
            "http://195.127.0.11/uploads/%20%20%20%20/.verify/.eBaysecure=updateuserdataxplimnbqmn-xplmvalidateinfoswqpcmlx=hgplmcx/",
        ),
        (
            "http://host%23.com/%257Ea%2521b%2540c%2523d%2524e%25f%255E00%252611%252A22%252833%252944_55%252B",
            "http://host%23.com/~a!b@c%23d$e%25f^00&11*22(33)44_55+",
        ),
        ("http://3279880203/blah", "http://195.127.0.11/blah"),
        ("http://www.google.com/blah/..", "http://www.google.com/"),
        ("http://a.com/../b", "http://a.com/b"),
        ("www.google.com/", "http://www.google.com/"),
        ("www.google.com", "http://www.google.com/"),
        ("http://www.evil.com/blah#frag", "http://www.evil.com/blah"),
        ("http://www.GOOgle.com/", "http://www.google.com/"),
        ("http://www.google.com.../", "http://www.google.com/"),
        ("http://www.google.com/foo\tbar\rbaz\n2", "http://www.google.com/foobarbaz2"),
        ("http://www.google.com/q?", "http://www.google.com/q?"),
        ("http://www.google.com/q?r?", "http://www.google.com/q?r?"),
        ("http://www.google.com/q?r?s", "http://www.google.com/q?r?s"),
        ("http://evil.com/foo#bar#baz", "http://evil.com/foo"),
        ("http://evil.com/foo;", "http://evil.com/foo;"),
        ("http://evil.com/foo?bar;", "http://evil.com/foo?bar;"),
        (b"http://\x01\x80.com/", "http://%01%80.com/"),
        ("http://notrailingslash.com", "http://notrailingslash.com/"),
        ("http://www.gotaport.com:1234/", "http://www.gotaport.com:1234/"),
        ("http://www.google.com:443/", "http://www.google.com:443/"),
        ("  http://www.google.com/  ", "http://www.google.com/"),
        ("http:// leadingspace.com/", "http://%20leadingspace.com/"),
        ("http://%20leadingspace.com/", "http://%20leadingspace.com/"),
        ("%20leadingspace.com/", "http://%20leadingspace.com/"),
        ("https://www.securesite.com:443/", "https://www.securesite.com/"),
        ("ftp://ftp.myfiles.com:21/", "ftp://ftp.myfiles.com/"),
        # ("http://some%1Bhost.com/%1B", "http://some%1Bhost.com/%1B"),
        # NOTE: Unlike Google, we ensure that the host name is always lower-cased.
        ("http://some%1Bhost.com/%1B", "http://some%1bhost.com/%1B"),
        # Test NULL character
        (b"http://test%00\x00.com/", "http://test%00%00.com/"),
        # Username and password should be removed
        # ("http://user:password@google.com/", "http://google.com/"),
        # NOTE: Unlike Google, we keep the username and password.
        ("http://user:password@google.com/", "http://user:password@google.com/"),
        # All of these cases are missing a valid hostname and should return ''
        ("", ""),
        (":", ""),
        ("/blah", ""),
        ("#ref", ""),
        ("/blah#ref", ""),
        ("?query#ref", ""),
        ("/blah?query#ref", ""),
        ("/blah;param", ""),
        ("http://#ref", ""),
        ("http:///blah#ref", ""),
        ("http://?query#ref", ""),
        ("http:///blah?query#ref", ""),
        ("http:///blah;param", ""),
        ("http:///blah;param?query#ref", ""),
        ("mailto:bryner@google.com", ""),
        # If the protocol is unrecognized, the URL class does not parse out
        # a hostname.
        # ("myprotocol://site.com/", ""),
        # This URL should _not_ have hostname shortening applied to it.
        (
            "http://i.have.way.too.many.dots.com/",
            "http://i.have.way.too.many.dots.com/",
        ),
        # WholeSecurity escapes parts of the scheme
        (
            "http%3A%2F%2Fwackyurl.com:80/",  # spell-checker: disable-line
            "http://wackyurl.com/",
        ),
        ("http://W!eird<>Ho$^.com/", "http://w!eird<>ho$^.com/"),
        # The path should have a leading '/' even if the hostname was terminated
        # by something other than a '/'.
        ("ftp://host.com?q", "ftp://host.com/?q"),
    ]
    for test, want in items:
        have = canonical_url(test)
        assert have == want, f"test input: {test!r}, actual: {have}, expected: {want!r}"


def test_url_() -> None:
    """Run Google's tests.

    See: https://developers.google.com/safe-browsing/v4/urls-hashing#canonicalization
    """
    Canonicalize = canonical_url
    assert Canonicalize("http://host/%25%32%35") == "http://host/%25"
    assert Canonicalize("http://host/%25%32%35%25%32%35") == "http://host/%25%25"
    assert Canonicalize("http://host/%2525252525252525") == "http://host/%25"
    assert Canonicalize("http://host/asdf%25%32%35asd") == "http://host/asdf%25asd"
    assert (
        Canonicalize("http://host/%%%25%32%35asd%%") == "http://host/%25%25%25asd%25%25"
    )
    assert Canonicalize("http://www.google.com/") == "http://www.google.com/"
    assert (
        Canonicalize(
            "http://%31%36%38%2e%31%38%38%2e%39%39%2e%32%36/%2E%73%65%63%75%72%65/%77%77%77%2E%65%62%61%79%2E%63%6F%6D/"
        )
        == "http://168.188.99.26/.secure/www.ebay.com/"
    )
    assert (
        Canonicalize(
            "http://195.127.0.11/uploads/%20%20%20%20/.verify/.eBaysecure=updateuserdataxplimnbqmn-xplmvalidateinfoswqpcmlx=hgplmcx/"
        )
        == "http://195.127.0.11/uploads/%20%20%20%20/.verify/.eBaysecure=updateuserdataxplimnbqmn-xplmvalidateinfoswqpcmlx=hgplmcx/"
    )
    assert (
        Canonicalize(
            "http://host%23.com/%257Ea%2521b%2540c%2523d%2524e%25f%255E00%252611%252A22%252833%252944_55%252B"
        )
        == "http://host%23.com/~a!b@c%23d$e%25f^00&11*22(33)44_55+"
    )
    assert Canonicalize("http://3279880203/blah") == "http://195.127.0.11/blah"
    assert Canonicalize("http://www.google.com/blah/..") == "http://www.google.com/"
    assert Canonicalize("www.google.com/") == "http://www.google.com/"
    assert Canonicalize("www.google.com") == "http://www.google.com/"
    assert Canonicalize("http://www.evil.com/blah#frag") == "http://www.evil.com/blah"
    assert Canonicalize("http://www.GOOgle.com/") == "http://www.google.com/"
    assert Canonicalize("http://www.google.com.../") == "http://www.google.com/"
    assert (
        Canonicalize("http://www.google.com/foo\tbar\rbaz\n2")
        == "http://www.google.com/foobarbaz2"
    )
    assert Canonicalize("http://www.google.com/q?") == "http://www.google.com/q?"
    assert Canonicalize("http://www.google.com/q?r?") == "http://www.google.com/q?r?"
    assert Canonicalize("http://www.google.com/q?r?s") == "http://www.google.com/q?r?s"
    assert Canonicalize("http://evil.com/foo#bar#baz") == "http://evil.com/foo"
    assert Canonicalize("http://evil.com/foo;") == "http://evil.com/foo;"
    assert Canonicalize("http://evil.com/foo?bar;") == "http://evil.com/foo?bar;"
    assert Canonicalize(b"http://\x01\x80.com/") == "http://%01%80.com/"
    assert Canonicalize("http://notrailingslash.com") == "http://notrailingslash.com/"
    # NOTE: Unlike Google, we do not remove non-default ports.
    # assert Canonicalize("http://www.gotaport.com:1234/") == "http://www.gotaport.com/"
    assert Canonicalize("  http://www.google.com/  ") == "http://www.google.com/"
    assert Canonicalize("http:// leadingspace.com/") == "http://%20leadingspace.com/"
    assert Canonicalize("http://%20leadingspace.com/") == "http://%20leadingspace.com/"
    assert Canonicalize("%20leadingspace.com/") == "http://%20leadingspace.com/"
    assert Canonicalize("https://www.securesite.com/") == "https://www.securesite.com/"
    assert Canonicalize("http://host.com/ab%23cd") == "http://host.com/ab%23cd"
    assert (
        Canonicalize("http://host.com//twoslashes?more//slashes")
        == "http://host.com/twoslashes?more//slashes"
    )



"""Ping and scrape links."""

# std
from time import time
from typing import Iterator
from typing import Literal
from typing import Optional
from typing import Tuple
from urllib.parse import urljoin
import re

# lib
import requests
from requests import Response
from bs4 import BeautifulSoup

# pkg
from .cache import LinkEntry
from .canonicalize import canonical_url

HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"
    )
}
"""HTTP headers to send when pinging."""

CSS_IMPORT = re.compile(r"""@import ["']([^"']+)""", re.MULTILINE)
"""Match a CSS `@import` expression."""

CSS_URL = re.compile(r"url\s*\(\s*([^)]+)\s*\)", re.MULTILINE)
"""Match a CSS `url()` expression."""

LINK_ATTRS = ["href", "src"]
"""HTML attributes that contain a URL."""

LINK_SELECT = ",".join(f"[{a}]" for a in LINK_ATTRS)
"""Query for selecting HTML tags that have URLs."""

LINK_IGNORE = re.compile(r"^(#|data:|mailto:|tel:)")
"""Pattern of links to ignore."""


def ping_url(
    url: str, method: Literal["head", "get"] = "get"
) -> Tuple[Optional[Response], LinkEntry]:
    """Make a request and return the response (if any) and `LinkEntry`."""
    t = int(time())
    try:
        res = getattr(requests, method)(url, timeout=3, headers=HEADERS)
        return res, LinkEntry(time=t, code=res.status_code, url=url)
    except Exception as e:
        return None, LinkEntry(time=t, code=-1, err=str(e), url=url)


def css_links(text: str) -> Iterator[str]:
    """Yield the links in a CSS document."""
    for pattern in [CSS_IMPORT, CSS_URL]:
        for link in pattern.findall(text):
            yield link


def html_links(soup: BeautifulSoup) -> Iterator[str]:
    """Yield the links in an HTML document."""
    for tag in soup.select(LINK_SELECT):
        if tag.name == "link" and "preconnect" in tag.get_attribute_list("rel", []):
            continue  # skip non-URL connection hints

        for attr in LINK_ATTRS:
            if attr in tag.attrs:
                yield tag.get_attribute_list(attr, [""])[0]


def scrape_links(res: Response) -> Iterator[str]:
    """Yield links in a `Response` based on content-type."""
    base = res.url
    ctype = res.headers["Content-Type"]
    links: Iterator[str] = iter([])

    if not ctype.startswith("text/"):  # skip non-textual (images, zip files)
        pass
    elif ctype.startswith("text/css") or base.endswith(".css"):
        links = css_links(res.text)
    elif ctype.startswith("text/xml") or base.endswith(".xml"):
        links = html_links(BeautifulSoup(res.content, features="xml"))
    else:  # assume HTML
        soup = BeautifulSoup(res.content, features="html.parser")
        links = html_links(soup)
        for tag in soup.find_all("base"):
            if "href" in tag.attrs:
                base = tag["href"]
                break  # only the first href counts

    for link in links:
        if not LINK_IGNORE.match(link):
            yield canonical_url(urljoin(base, link))



#!/usr/bin/env python
"""Check to make sure your links are still fresh.

```text
Usage: freshlinks
    [--help | --version] [--debug]
    [--config PATH]
    [--cache DIR] [--cache-age AGE] [--cache-ignore DOMAIN...] [--cache-retry]
    [<url>...] [--ignore URL...]
    [--domain DOMAIN...] [--depth NUM]
    [--show-ok] [--show-err]

Options:
  -h, --help                show this message and exit
  --version                 show version and exit
  --debug                   show debug messages

Config:
  -c PATH, --config PATH    path to configuration file

Cache:
  --cache DIR               directory to load and store cache files
                            [default: ~/.cache/freshlinks]

  --cache-age AGE           ignore cache data older than this age [default: 5d]
                            units: d=days; h=hours; m = minutes; s=sec

  --cache-ignore DOMAIN     ignore cache for these domains (always ping)
  --cache-retry             ignore cached result if it was an error

Check:
  <url>                     URL to check
  --ignore URL              URL to ignore

Scrape:
  --domain DOMAIN           domains to include in link checking
  --depth NUM               maximum recursion depth for checking links [default: 10]

Output:
  --show-ok                 show URLs that were successfully reached
  --show-err                show error messages

Examples:
$ python -m freshlinks http://example.com
```
"""

# std
from __future__ import annotations
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from random import uniform
from time import sleep
from time import time
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Set
from typing import Tuple

# lib
from attrbox import AttrDict
from attrbox import load_config
from attrbox import parse_docopt
from multiprocess import Manager  # type: ignore
from url_normalize.tools import deconstruct_url  # type: ignore
import ezq

# pkg
from .cache import age_time
from .cache import LinkCache
from .cache import LinkEntry
from .canonicalize import canonical_url
from .scraper import ping_url
from .scraper import scrape_links
from . import __version__


@dataclass
class Args:
    help: bool = False
    version: bool = False
    debug: bool = False
    config: Optional[Path] = None
    cache: Optional[Path] = None
    cache_age: str = "5d"
    cache_ignore: Set[str] = field(default_factory=set)
    cache_retry: bool = False
    url: List[str] = field(default_factory=list)
    ignore: List[str] = field(default_factory=list)
    domain: Set[str] = field(default_factory=set)
    depth: int = 10
    show_ok: bool = False
    show_err: bool = False

    @staticmethod
    def load() -> Args:
        """Load CLI arguments and any configuration files."""
        args = parse_docopt(__doc__, version=__version__, read_config=False)
        config = AttrDict(load_config(Path(args.config))) if args.config else AttrDict()

        # NOTE: args overwrite configs
        for arg, val in args.items():
            if arg not in config:  # config doesn't have this arg
                config[arg] = val
            elif isinstance(val, list):  # merge lists
                config[arg] = list(config[arg] or []) + val
            elif isinstance(val, bool):  # merge bools
                config[arg] = config[arg] or val
            else:  # by default, arg wins
                config[arg] = val

        config.config = Path(config.config) if config.config else None
        config.cache = Path(config.cache) if config.cache else None
        config.cache_ignore = set(config.cache_ignore or [])
        config.url = [canonical_url(url) for url in config.url] if config.url else []
        config.depth = int(config.depth) if config.depth is not None else 10
        config.domain = set(
            config.domain or {deconstruct_url(url).host for url in (config.url or [])}
        )

        return Args(**config)


def inc_keys(stat: Dict[str, int], *keys: str) -> None:
    """Increment the string keys of a dict by 1."""
    for key in keys:
        stat[key] = stat.get(key, 0) + 1


def print_stats(stat: Dict[str, int]) -> None:
    """Print right-aligned sorted keys with values."""
    print("")
    n = max(len(k) for k in stat.keys())
    for k, v in sorted(stat.items()):
        print(f"{k: >{n}}:", v)


def print_entry(args: Args, entry: LinkEntry) -> None:
    """Print the result of a ping."""
    if entry.ok and not args.show_ok:
        return

    print(
        f"{'ERR' if entry.code == -1 else entry.code} "
        f"{'(cache) ' if entry.is_cached else ''}"
        f"{entry.url}"
    )
    if args.show_err:
        print(f" ==> {entry.err}")


def process_link(
    url: str,
    depth: int,
    args: Args,
    todo: List[Tuple[str, int]],
    done: Dict[str, Optional[LinkEntry]],
    stat: Dict[str, int],
    cache: LinkCache,
) -> None:
    domain = deconstruct_url(url).host
    do_scrape = domain in args.domain and depth <= args.depth

    entry = None
    if depth > 0 and domain not in args.cache_ignore and cache and url in cache:
        entry = cache[url]
        if entry:
            inc_keys(stat, "cache")

    if not entry or (args.cache_retry and not entry.ok):
        method: Literal["get", "head"] = "get" if do_scrape else "head"
        req, entry = ping_url(url, method)
        inc_keys(stat, f"ping:{method}")

        if do_scrape and req:
            entry.links = list(scrape_links(req))
            inc_keys(stat, "scrape")

        done[url] = entry
        if cache:
            cache[url] = entry

    inc_keys(stat, "total", str(entry.code), "ok" if entry.ok else "fail")
    print_entry(args, entry)

    if do_scrape and url not in args.ignore:
        for link in entry.links:
            if link not in done and link not in args.ignore:
                todo.append((link, depth + 1))


def worker_print(n: int, t: float, args: Args, msg: str) -> None:
    """Print a message from a worker."""
    if not args.debug:
        return
    print(f"{time() - t:07.1f} WORKER-{n}:", msg)


def worker_sleep(n: int, t: float, args: Args, todo: List[Tuple[str, int]]) -> None:
    """Sleep a worker, if needed."""
    if not todo:
        worker_print(n, t, args, "sleep")
        sleep(uniform(1, 4))  # wait for some more work
        worker_print(n, t, args, "awake")


def worker(
    n: int,
    args: Args,
    todo: List[Tuple[str, int]],
    done: Dict[str, Optional[LinkEntry]],
    stat: Dict[str, int],
    cache: LinkCache,
) -> None:
    """Run a worker."""
    t = time()
    worker_print(n, t, args, "beg")
    worker_sleep(n, t, args, todo)  # don't start too early
    while todo:
        url, depth = None, 0
        try:
            url, depth = todo.pop()
        except IndexError:  # todo might actually be empty
            pass

        if url and url not in done:  # new link
            done[url] = None  # prevent other workers
            process_link(url, depth, args, todo, done, stat, cache)

        worker_sleep(n, t, args, todo)  # don't end too early
    worker_print(n, t, args, "end")


def main() -> None:
    """Main entry point."""
    t = time()
    args = Args.load()
    if args.debug:
        print(args)

    with Manager() as manager:
        todo = manager.list([(url, 0) for url in args.url])
        done = manager.dict()
        stat = manager.dict()
        cache = (
            LinkCache(
                manager.dict(),
                max_time=age_time(args.cache_age),
                path=args.cache,
            )
            if args.cache
            else None
        )

        _bundle = (args, todo, done, stat, cache)
        for w in [ezq.run(worker, n, *_bundle) for n in range(ezq.NUM_CPUS)]:
            w.join()

        stat["time"] = f"{time() - t:2.1f}s"
        print_stats(stat)


if __name__ == "__main__":
    main()



"""Generate canonical URLs roughly using Google's Safe Browsing standard.

Differences:

- Fragments are removed, but `#!` is converted to `?_escaped_fragment_=`
- Username and password are NOT removed from the host.

See:

- [URLs and Hashing: Canonicalization](https://developers.google.com/safe-browsing/v4/urls-hashing#canonicalization)
- [`expression.py`](https://chromium.googlesource.com/external/google-safe-browsing/+/06a8c4e799233da220ad7411e2bfacc74cbfbb37/python/expression.py)
- [`url_normalize.py`](https://github.com/niksite/url-normalize/blob/master/url_normalize/url_normalize.py)
"""

# This file is a rewrite of Google's Python 2.5 implementation.
#
# Copyright 2023 Metaist LLC.
# Licensed under the MIT License.
#
# The original `expression.py` file carries this license:
#
# Copyright 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# std
from typing import List
from typing import Union
from typing import cast
from urllib.parse import quote
from urllib.parse import unquote
from urllib.parse import unquote_to_bytes
import re
import string

# lib
from attrbox import AttrDict
from url_normalize.tools import deconstruct_url  # type: ignore
from url_normalize.tools import reconstruct_url
from url_normalize.url_normalize import normalize_port  # type: ignore
from url_normalize.url_normalize import normalize_query
from url_normalize.url_normalize import normalize_userinfo
from url_normalize.url_normalize import provide_url_scheme

DEFAULT_SCHEME = "http"
"""Default scheme that browsers uses."""

SAFE_CHARS = "".join(
    c
    for c in string.digits + string.ascii_letters + string.punctuation
    if c not in "%#"
)
"""Characters that are not escaped."""

IP_WITH_TRAILING_SPACE = re.compile(r"^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) ")
POSSIBLE_IP = re.compile(r"^((?:0x[0-9a-f]+|[0-9\\.])+)$", flags=re.I)
FIND_BAD_OCTAL_REGEXP = re.compile(r"(^|\.)0\d*[89]")

HEX = re.compile(r"^0x([a-fA-F0-9]+)$")
OCT = re.compile(r"^0([0-7]+)$")
DEC = re.compile(r"^(\d+)$")


def unquote2(string: Union[str, bytes]) -> Union[str, bytes]:
    """Unquote both strings and bytes."""
    if isinstance(string, bytes):
        return unquote_to_bytes(string)
    return unquote(string)


def escape(string: Union[str, bytes]) -> str:
    """Fully escape `string`, then re-escape once."""
    # See: https://chromium.googlesource.com/external/google-safe-browsing/+/06a8c4e799233da220ad7411e2bfacc74cbfbb37/python/expression.py#292
    unquoted = unquote2(string)
    while unquoted != string:
        string = unquoted
        unquoted = unquote2(unquoted)
    return quote(unquoted, SAFE_CHARS)


def canonical_ip(host: str) -> str:
    """Return a canonical IP address."""
    if len(host) <= 15:
        # This handles the Windows resolver allows an IP
        # followed by a space and something else as long as it
        # is under 15 characters.
        if m := IP_WITH_TRAILING_SPACE.match(host):
            host = m.group(1)

    if not POSSIBLE_IP.match(host):
        return ""

    # Try to parse octal, if possible.
    allow_octal = not FIND_BAD_OCTAL_REGEXP.search(host)

    # Skip trailing, leading and consecutive dots.
    parts = [part for part in host.split(".") if part]
    if len(parts) > 4:
        return ""

    ip: List[str] = []
    for i, part in enumerate(parts):
        if m := HEX.match(part):
            base = 16
        elif allow_octal and (m := OCT.match(part)):
            base = 8
        elif m := DEC.match(part):
            base = 10
        else:
            return ""

        # print("part:", part, "m:", m.group(1), "base:", base)
        n = int(m.group(1), base)
        if n <= 255:
            ip.append(str(n))
            continue

        # print("n > 255:", n)
        if i < len(parts) - 1:
            n &= 0xFF
            ip.append(str(n))
        else:
            bar = bytearray()
            while n > 0 and len(bar) < 4:
                bar.append(n & 0xFF)
                n >>= 8

            if len(ip) + len(bar) > 4:
                return ""

            bar.reverse()
            ip.extend(str(b) for b in bar)

    return ".".join((ip + (["0"] * 4))[:4])


def canonical_host(host: str) -> str:
    """Return a canonical hostname."""
    # 0: IDN host names should be converted to ASCII punycode
    # 1: Remove all leading and trailing dots.
    # 2: Replace consecutive dots with a single dot.
    # 3: If the hostname can be parsed as an IP address, normalize it to
    #    4 dot-separated decimal values. The client should handle any legal IP-address
    #    encoding, including octal, hex, and fewer than four components.
    # 4: Lowercase the whole string.

    # See: https://chromium.googlesource.com/external/google-safe-browsing/+/06a8c4e799233da220ad7411e2bfacc74cbfbb37/python/expression.py#207
    if not host:
        return ""

    result = host.lower()  # Rule 4
    result = ".".join([part for part in result.split(".") if part])  # Rule 1 & 2
    result = result.encode("idna").decode("utf-8")  # Rule 0

    if ip := canonical_ip(host):  # Rule 3
        return ip

    return result


def canonical_path(path: str) -> str:
    """Return a canonical path."""
    # 1: Resolve the sequences "/../" and "/./" in the path by replacing "/./" with "/",
    #    and removing "/../" along with the preceding path component.
    # 2: Replace runs of consecutive slashes with a single slash character.

    # See: https://chromium.googlesource.com/external/google-safe-browsing/+/06a8c4e799233da220ad7411e2bfacc74cbfbb37/python/expression.py#157
    result = path
    if not result:
        return "/"

    if result[0] != "/":
        result = f"/{result}"

    result = escape(result)
    parts: List[str] = []
    for part in result.split("/"):
        if part == "..":  # remove previous part (if any)
            if len(parts) > 0:
                parts.pop()
        elif part and part != ".":  # skip empty and .
            parts.append(part)

    result = f"/{'/'.join(parts)}"
    if path.endswith("/") and not result.endswith("/"):
        result = f"{result}/"
    # leading and trailing slashes added (if needed)

    # SPECIAL CASE: Handle URLs tacked on.
    result = result.replace("http:/", "http://").replace("https:/", "https://")
    return result


def canonical_url(url: Union[str, bytes]) -> str:
    """Return a canonical version of `url`."""
    # See: https://developers.google.com/safe-browsing/v4/urls-hashing#canonicalization
    # See: https://chromium.googlesource.com/external/google-safe-browsing/+/06a8c4e799233da220ad7411e2bfacc74cbfbb37/python/expression.py
    # 4: Remove tab (0x09), CR (0x0d), and LF (0x0a) characters from the URL.
    #    Do not remove escape sequences for these characters (e.g. '%0a').

    # 1: URL is valid RFC 2396
    # 2: Convert internationalized domain name (IDN) to ASCII Punycode.
    # 3: URL must include a path.
    # 5: Remove the fragment.
    # 6: Repeatedly percent-unescape the URL until it has no more percent-escapes.
    url = url.strip()  # no leading or trailing whitespace

    has_end_q = False
    if isinstance(url, bytes):
        url = url.replace(b"\t", b"").replace(b"\r", b"").replace(b"\n", b"")  # Rule 4
        url = url.replace(b"#!", b"?_escaped_fragment_=")  # Different than Google
        if (pos := url.find(b"#")) >= 0:
            url = url[0:pos]  # Rule 5
        has_end_q = url.endswith(b"?")
    else:
        url = url.replace("\t", "").replace("\r", "").replace("\n", "")  # Rule 4
        url = url.replace("#!", "?_escaped_fragment_=")  # Different than Google
        if (pos := url.find("#")) >= 0:
            url = url[0:pos]  # Rule 5
        has_end_q = url.endswith("?")

    url = escape(url)  # in case, e.g., scheme has encoded characters
    if url.startswith(":") and len(url) > 1:
        url = f"localhost{url}"
    url = provide_url_scheme(url, DEFAULT_SCHEME)

    parts = AttrDict(deconstruct_url(url)._asdict())
    if not parts.host:
        return ""

    parts.scheme = parts.scheme.lower() if parts.scheme else DEFAULT_SCHEME
    parts.userinfo = normalize_userinfo(parts.userinfo)
    parts.host = canonical_host(parts.host)
    parts.port = normalize_port(parts.port, parts.scheme)
    parts.path = canonical_path(cast(str, parts.path))
    parts.query = normalize_query(parts.query)
    parts.fragment = ""

    result = str(reconstruct_url(parts))
    if has_end_q and not result.endswith("?"):
        result += "?"

    return result



"""Test canonical components beyond Google's test suite."""

# pkg
from freshlinks.canonicalize import canonical_host
from freshlinks.canonicalize import canonical_ip
from freshlinks.canonicalize import canonical_path
from freshlinks.canonicalize import canonical_url


def test_ip_overflow() -> None:
    """High IP components."""
    assert canonical_ip("256.0.0.0") == "0.0.0.0"
    assert canonical_ip("300.0.0.0") == "44.0.0.0"
    assert canonical_ip("0.0.0.256") == "", "overflow of last component"


def test_empty_host() -> None:
    """Empty host name."""
    assert canonical_host("") == ""
    assert canonical_url(":8080") == "http://localhost:8080/"


def test_empty_path() -> None:
    """Empty path."""
    assert canonical_path("") == "/"
    assert canonical_path("foo") == "/foo"


def test_way_back_machine_path() -> None:
    """WayBackMachine path."""
    url = "https://web.archive.org/web/20230609101345/https://metaist.com/blog/"
    assert canonical_url(url) == url


def test_bytes_url_fragment() -> None:
    """Bytes URL"""
    assert canonical_url(b"http://example.com#foo") == "http://example.com/"
    assert (
        canonical_url(b"http://example.com#!foo")
        == "http://example.com/?_escaped_fragment_=foo"
    )



"""Cache URL response."""

# native
from __future__ import annotations
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from datetime import timedelta
from hashlib import sha256
from pathlib import Path
from time import time
from typing import Dict
from typing import Any
from typing import List
from typing import Optional
from typing import Union
import re

# lib
import ujson

StrOrPath = Union[str, Path]
"""`str` or `Path`."""

DEFAULT_PATH: StrOrPath = "~/.cache/freshlinks"
"""Default cache folder."""

RE_DURATION = re.compile(r"(\d+(?:\.\d+)?)[ ]*([dhms])")
"""Pattern for time strings."""

DURATION_UNITS = {
    "d": "days",
    "h": "hours",
    "m": "minutes",
    "s": "seconds",
}
"""Time string units."""


def to_timedelta(duration: str) -> timedelta:
    """Convert a time duration string into a `timedelta`."""
    parts = re.findall(RE_DURATION, duration.lower())
    args = {"days": 0.0, "hours": 0.0, "minutes": 0.0, "seconds": 0.0}
    for value, unit in parts:
        args[DURATION_UNITS.get(unit, "seconds")] += float(value)
    return timedelta(**args)


def age_time(age: str, start: Optional[datetime] = None) -> int:
    """Return unix time relative to a start date.

    Args:
        age (str): amount of time; a number followed by a time unit
            (s=seconds, m=minutes, h=hours, d=days). `s` is default.

        start (datetime, optional): start date from which to subtract `age`.
            If `None` uses `datetime.now()`. Defaults to `None`.

    Returns:
        int: unix time relative to `start`
    """
    return int(((start or datetime.now()) - to_timedelta(age)).timestamp())


def hash_url(url: str) -> str:
    """Return the hash for a given `url`."""
    return sha256(url.encode("utf-8")).hexdigest()


@dataclass
class LinkEntry:
    is_cached: bool = False
    """Did this entry come from a cache?"""

    version: int = 1
    """Entry format version."""

    time: int = 0
    code: int = 200
    err: str = ""
    url: str = ""
    links: List[str] = field(default_factory=list)

    @staticmethod
    def load(data: Dict[str, Any]) -> LinkEntry:
        return LinkEntry(
            is_cached=True,
            version=data["v"],
            time=data["t"],
            code=data["c"],
            err=data.get("err", ""),
            url=data["url"],
            links=data["links"],
        )

    @property
    def ok(self) -> bool:
        return self.code in [200, 301, 302]

    @property
    def json(self) -> Dict[str, Any]:
        return dict(
            v=self.version,
            t=self.time,
            c=self.code,
            err=self.err,
            url=self.url,
            links=self.links,
        )


class LinkCache:
    """A file-based cache for storing metadata about URLs."""

    path: Path
    """Cache folder."""

    max_time: int = 0
    """Cache entries older than this are considered invalid."""

    def __init__(
        self,
        store: Dict[str, LinkEntry],
        max_time: int = 0,
        path: StrOrPath = DEFAULT_PATH,
        preload: bool = False,
    ):
        self.store = store
        self.max_time = max_time if max_time > 0 else int(time())
        self.path = Path(path).expanduser()
        self.path.mkdir(exist_ok=True)  # ensure cache folder exists
        if preload:
            self.preload()

    def preload(self) -> None:
        for item in self.path.glob("*.json"):
            self.load(item.stem)

    def load(self, hashed: str) -> Optional[LinkEntry]:
        path = self.path / f"{hashed}.json"
        if not path.exists():
            return None

        data = ujson.loads(path.read_text())
        if data.get("t", 0) < self.max_time:
            return None

        entry = LinkEntry.load(data)
        self.store[entry.url] = entry
        return entry

    def __contains__(self, url: str) -> bool:
        return url in self.store or (self.path / f"{hash_url(url)}.json").exists()

    def __getitem__(self, url: str) -> Optional[LinkEntry]:
        try:
            return self.store[url]
        except KeyError:
            return self.load(hash_url(url))

    def __setitem__(self, url: str, entry: LinkEntry) -> None:
        self.store[url] = entry
        if entry:
            (self.path / f"{hash_url(url)}.json").write_text(
                ujson.dumps(
                    entry.json,
                    ensure_ascii=False,
                    escape_forward_slashes=False,
                )
            )
